from tornado.options import options, define
from .json_utils import dumps
import asyncio, aioredis
import logging
import redis


async def connect(redis_url: str = None):
    """
        Create a single connection to redis using
        option.redis_url if none is supplied
    """
    redis_url = options.redis_url if redis_url is None else redis_url
    loop = asyncio.get_event_loop()
    redis = await aioredis.create_redis(redis_url, loop=loop)
    logging.info("connect to redis: %s", redis_url)
    return redis


async def pool(redis_url: str = None):
    """
        Create a pool connection to redis using
        option.redis_url if none is supplied
    """
    redis_url = options.redis_url if redis_url is None else redis_url
    loop = asyncio.get_event_loop()
    pool = await aioredis.create_pool(redis_url, minsize=5, maxsize=10, loop=loop)
    logging.info("pool to redis: %s", redis_url)
    return pool



"""
    A service is a task that runs in the background.
    It can be started and stopped.
"""
import asyncio
import logging

LOGGER = logging.getLogger(__name__)


class ServiceManager:
    """ This service manager only support async functions """

    def __init__(self, services):
        self._services = {
            srv.__name__: srv for srv in self._discover_(services)
        }
        self._tasks = {}
        LOGGER.debug("services %s", self.available_services)
        asyncio.ensure_future(self._run_all_())

    @classmethod
    def _discover_(cls, services):
        """ utility to discover public functions in a module """
        for name in dir(services):
            if name[0] == "_":
                continue
            service = getattr(services, name)
            if callable(service):
                yield service

    @property
    def available_services(self):
        """ names of available services """
        return [name for name in self._services]

    async def run_service(self, name: str):
        """ create a task of a function """
        if name not in self._tasks:
            LOGGER.info("running service %s", name)
            task = asyncio.ensure_future(self._services[name]())
            self._tasks[name] = task

    async def stop_service(self, name: str):
        """ stop a task """
        if name in self._tasks:
            LOGGER.info("stopping service %s", name)
            task = self._tasks[name]
            task.cancel()
            del self._tasks[name]

    async def _run_all_(self):
        """ helper to run all """
        for name in self.available_services:
            await self.run_service(name)

    async def _stop_all_(self):
        """ helper to shut down """
        for name in self._tasks:
            await self.stop_service(name)



from tornado.ioloop import IOLoop
from tornado import gen
import asyncio
import threading
import functools
import logging

LOGGER = logging.getLogger(__name__)


_context_ = threading.local()
_manager_ = None


class Context:
    """
        Used to capture broadcast
        that should only fly if
        the action completes.
    """

    def __init__(self, result, messages):
        self.result = result
        self.messages = messages

    def to_json(self):
        return {
            "_type_": self.__class__.__name__,
            "result": self.result,
            "messages": self.messages,
        }


def get_current_user():
    if hasattr(_context_, "user"):
        return _context_.user


def set_current_user(user):
    LOGGER.info("current user: %s", user)
    _context_.user = user


def broadcast(signal, message, filter_clients=None):
    _manager_.broadcast(signal, message, filter_clients)


def broadcast_on_success(signal, message, filter_clients=None):
    if hasattr(_context_, "broadcasts"):
        _context_.broadcasts.append((signal, message, filter_clients))
    else:
        broadcast(signal, message, filter_clients)


def register_object(obj: object) -> str:
    return _manager_.register_state(obj)


def deregister_object(name: str):
    return _manager_.deregister_state(name)


def put_work(actor, method, params):
    content = {"current_user": actor, "method": method, "params": params}
    _manager_.ioloop.add_callback(_manager_.put, content)


async def perform(actor, fn, args, kwargs):
    ioloop = IOLoop.current()
    if asyncio.iscoroutinefunction(fn):
        result = asyncio.ensure_future(fn(*args, **kwargs))
    elif gen.is_coroutine_function(fn):
        result = fn(*args, **kwargs)
    else:
        result = ioloop.run_in_executor(None, _perform_, actor, fn, args, kwargs)
    if asyncio.isfuture(result):
        result = await result
    return result


def _perform_(actor, fn, args, kwargs):
    _context_.broadcasts = []
    set_current_user(actor)

    try:
        LOGGER.info("%s %r %r", fn, args, kwargs)
        result = fn(*args, **kwargs)
        result = Context(result, _context_.broadcasts)
        return result
    except Exception as ex:
        LOGGER.exception(ex)
        raise ex
    finally:
        _context_.broadcasts = []
        _context_.user = None



import argparse
import logging
import os
import signal


logger = logging.getLogger(__name__)


class Stoppable:
    def __init__(self):
        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            description="script to control the process that"
            " runs the threads in the background."
        )
        self.parser.add_argument(
            "command",
            metavar="command",
            help="the command to be run, may be `stop` or `start`",
        )
        self.parser.add_argument(
            "-p",
            "--pid",
            dest="pid_file",
            required=True,
            metavar="/path/to/pid/file",
            help="the path to where the pid file will be / is stored",
        )
        self.args = None

    def write_pid(self):
        pid = os.getpid()
        with open(self.args.pid_file, "w") as f:
            f.write(f"{pid}")

    def _start(self):
        self.write_pid()
        self.start()

    def start(self):
        pass

    def stop(self):
        pid_file = self.args.pid_file
        try:
            with open(pid_file, "r") as f:
                contents = f.read()
            pid = int(contents)
            os.kill(pid, signal.SIGINT)
            os.remove(pid_file)
        except:
            logger.info("could not find %s", pid_file)

    def main(self):
        self.args = self.parser.parse_args()
        if self.args.command == "start":
            self._start()
        elif self.args.command == "stop":
            self.stop()



from tornado.options import options
from tornado.queues import Queue
from tornado.concurrent import Future
from concurrent.futures import ThreadPoolExecutor
from collections import OrderedDict
from tornado import gen
from ..utils.status import Status
from ..utils import json_utils
from . import perform
import tornado.ioloop
import importlib
import inspect
import functools
import asyncio
import time
import logging

LOGGER = logging.getLogger(__name__)


class Manager:

    _manager_ = None

    def __init__(self, procedures=None, workers=2):
        self.clients = []
        self.ioloop = tornado.ioloop.IOLoop.current()
        self.status = Status(self.ioloop, self.publish_status, timeout=1)
        self.procedures = {}
        self.proc_names = []
        self.reflection = OrderedDict()
        if procedures:
            self.add_targets(procedures)
        self.stateful = {}

        # we're on the main thread
        perform._manager_ = self
        self._manager_ = self
        self._setup_queues_(workers)
        tornado.ioloop.PeriodicCallback(self.keep_alive, 30000).start()

    def add_targets(self, procedures):
        names = getattr(
            procedures, "__all__", [p for p in dir(procedures) if p[0] != "_"]
        )
        self.procedures.update({name: getattr(procedures, name) for name in names})
        self.proc_names.extend(names)
        self.reflection.update(self.describe(procedures, names))
        LOGGER.info("avail: %s", self.proc_names)

    def get_target(self, name):
        return self.procedures.get(name)

    def publish_status(self, status):
        self.broadcast("set_status", status)

    def keep_alive(self):
        """ used to ping the client """
        msg = str(time.time()).encode("utf8")
        LOGGER.debug("ping %s", msg)
        for client in self.clients:
            client.ping(msg)

    def register_state(self, obj):
        """
            Keeps a stateful object
        """
        name = str(id(obj))
        self.stateful[name] = obj
        return name

    def deregister_state(self, name: str):
        """
            Deletes a stateful object
        """
        del self.stateful[name]

    def _setup_queues_(self, workers):
        self._e = ThreadPoolExecutor()
        self._q = Queue()
        for _ in range(workers):
            self.ioloop.spawn_callback(self.work, self._q)

    def add_client(self, client):
        self.clients.append(client)
        self.status["clients"] = len(self.clients)
        client.write_message(
            json_utils.dumps({"signal": "set_reflect", "message": self.reflection})
        )

    def remove_client(self, client):
        if client in self.clients:
            self.clients.remove(client)
            self.status["clients"] = len(self.clients)

    def filter_client(self, client, filter_clients):
        """ Does this client meet these criteria """
        return True

    def _broadcast_(self, signal, message, filter_clients=None):
        data = json_utils.dumps({"signal": signal, "message": message})
        for client in self.clients:
            if self.filter_client(client, filter_clients):
                client.write_message(data)

    def broadcast(self, signal, message, filter_clients=None):
        """
            a thread safe way to broadcast
        """
        self.ioloop.spawn_callback(self._broadcast_, signal, message, filter_clients)

    def status_update(self):
        self.status["q"] = self._q.qsize()

    def get_params(self, content):
        if isinstance(content["params"], dict):
            return [], content["params"]
        elif isinstance(content["params"], list):
            return content["params"], {}
        raise Exception("Params neither list or dict")

    @gen.coroutine
    def put(self, content):
        future = Future()
        yield self._q.put((content, future))
        self.status_update()
        return future

    @gen.coroutine
    def work(self, q):
        while True:
            item = yield q.get()
            content, future = item
            try:
                method = content["method"]
                actor = content.get("current_user")
                args, kwargs = self.get_params(content)
                if ":" in method:
                    mod_name, func_name = method.split(":")
                    procedure = getattr(importlib.import_module(mod_name), func_name)
                else:
                    procedure = self.get_target(method)
                LOGGER.info("work on %s", procedure)
                result = yield perform.perform(actor, procedure, args, kwargs)
                if result and isinstance(result, perform.Context):
                    future.set_result(result.result)
                    for signal, message, filter_clients in result.messages:
                        self.broadcast(signal, message, filter_clients)
                else:
                    future.set_result(result)
            except Exception as ex:
                LOGGER.exception(ex)
                future.set_exception(ex)
            finally:
                q.task_done()
                self.status_update()

    @classmethod
    def _annotation_to_str_(cls, annotation):
        try:
            if annotation is None:
                return None
            if isinstance(annotation, str):
                return annotation
            if issubclass(annotation, object):
                return annotation.__name__
        except:
            # for some reason a Tuple was throwing and exception
            pass
        return str(annotation)

    @classmethod
    def describe(cls, target, labels=None):
        """
            Will dir target and inspect each
            function and return an instance of
            this class for each

            functions starting with '_' are ignored
        """
        result = OrderedDict()
        targets = labels if labels else dir(target)
        for key in filter(lambda x: x[0] != "_", targets):
            f = getattr(target, key)
            if inspect.isfunction(f) or inspect.ismethod(f):
                sig = inspect.signature(f)
                s = OrderedDict(
                    [
                        ("name", key),
                        ("params", [str(p) for p in sig.parameters.values()]),
                        (
                            "returns",
                            cls._annotation_to_str_(sig.return_annotation)
                            if sig.return_annotation is not sig.empty
                            else "",
                        ),
                        ("docs", inspect.getdoc(f)),
                        ("auth", hasattr(f, "auth")),
                    ]
                )
                LOGGER.debug(s)
                result[key] = s
            else:
                LOGGER.debug("not avaiable %r", f)
        return result



from contextlib import contextmanager
from tornado.options import options
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine import reflection
from sqlalchemy.sql.schema import MetaData, ForeignKeyConstraint, Table
from sqlalchemy.sql.ddl import DropConstraint, DropTable
import logging


class db:
    Session = None
    Engine = None
    url = None


Sessions = {}


def setup_db(base_name="default", db_url=None):
    logging.info("setUp(%r)", base_name)
    global Sessions
    if Sessions.get(base_name) is None:
        _db = db()
        _db.url = db_url if db_url else getattr(options, f"{base_name}_db_url")
        logging.info("connecting to db: %s", _db.url)
        _db.Session, _db.Engine = connect(_db.url)
        Sessions[base_name] = _db


def connect(db_url, echo=False):
    Engine = create_engine(db_url, echo=echo)
    Session = sessionmaker(bind=Engine)
    return Session, Engine


@contextmanager
def session(Session):
    """Provide a transactional scope around a series of operations."""
    session = Session()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


def session_scope(base_name="default", db_url=None):
    """Provide a transactional scope around a series of operations."""
    global Sessions
    if Sessions.get(base_name) is None:
        setup_db(base_name, db_url)
    return session(Sessions[base_name].Session)


def drop_all(session):
    """
        drops current schema
    """
    inspector = reflection.Inspector.from_engine(session.bind)

    # gather all data first before dropping anything.
    # some DBs lock after things have been dropped in
    # a transaction.

    metadata = MetaData()

    tbs = []
    all_fks = []

    for table_name in inspector.get_table_names():
        fks = []
        for fk in inspector.get_foreign_keys(table_name):
            if not fk["name"]:
                continue
            fks.append(ForeignKeyConstraint((), (), name=fk["name"]))
        t = Table(table_name, metadata, *fks)
        tbs.append(t)
        all_fks.extend(fks)

    for fkc in all_fks:
        session.execute(DropConstraint(fkc))

    for table in tbs:
        session.execute(DropTable(table))

    session.commit()



import asyncio
import logging
import time
from bluemax import context

LOGGER = logging.getLogger(__name__)

async def clock():
    while True:
        message = {"now": time.time()}
        LOGGER.debug(message)
        context.broadcast("time", message)
        await asyncio.sleep(5)



from tornado.web import RequestHandler, HTTPError
from tornado.options import options
from ...utils import json_utils
import asyncio
import logging

"""
    Support for Tornado Authentication.
"""


class UserMixin:
    """
        Should be mixed in with WebsocketHandlers
        or RequestHandlers. Make sure it is the
        first class in the declaration as it
        overrides methods in both handlers.
    """

    @property
    def cookie_name(self):
        """ return the cookie_name declared in options"""
        return options.cookie_name

    def get_current_user(self):
        """ return the current user from the cookie """
        result = self.get_secure_cookie(self.cookie_name)
        logging.debug("%s:%s", self.cookie_name, result)
        if result:
            result = json_utils.loads(result.decode("utf-8"))
        return result

    def set_current_user(self, value):
        """ put the current user in the cookie """
        if value:
            self.current_user = value
            self.set_secure_cookie(self.cookie_name, json_utils.dumps(value))
        else:
            self.clear_cookie(self.cookie_name)

    def set_transfer_user(self, value):
        return self.create_signed_value(
            "transfer-user", value=json_utils.dumps(value).encode("utf-8")
        )

    def get_transfer_user(self, value):
        user = self.get_secure_cookie("transfer-user", value)
        if user:
            self.current_user = user
            logging.info("current_user set: %s", user)
            return json_utils.loads(user.decode("utf-8"))


class LoginHandler(UserMixin, RequestHandler):
    """
        Can be called as ajax from the
        websocket client to get the auth cookie
        into the headers.
    """

    def login(self, username, password):
        raise Exception("subclass responsibility")

    async def post(self):
        username = self.get_argument("username")
        password = self.get_argument("password")

        user = self.login(username, password)
        if asyncio.iscoroutine(user):
            user = await user
        if user:
            self.set_current_user(user)
            self.write(self.set_transfer_user(user))
        else:
            self.set_status(403)
            self.finish("<html><body>Username or password incorrect</body></html>")


class LogoutHandler(UserMixin, RequestHandler):
    def get(self):
        """ removes cookie and redirects to optional next argument """
        self.set_current_user(None)
        self.redirect(self.get_argument("next", "/"))



from tornado.ioloop import IOLoop
from tornado.options import options
from tornado.platform.asyncio import AnyThreadEventLoopPolicy
from concurrent.futures import ThreadPoolExecutor
from tornado import autoreload
from ..web.settings import settings
from ..utils.rd_session import connect
from ..utils.json_utils import dumps, loads
from ..utils.config import extend
from . import perform
import importlib
import threading
import asyncio
import logging

LOGGER = logging.getLogger(__name__)

"""
    Main worker file. It subscribes to the
    redis queue and broadcasts activity
"""


class Worker:
    def __init__(self, ioloop):
        self.ioloop = ioloop
        self.ioloop.set_default_executor(ThreadPoolExecutor())
        self.procedures = importlib.import_module(options.procedures)

    def get_params(self, content):
        if isinstance(content["params"], dict):
            return [], content["params"]
        elif isinstance(content["params"], list):
            return content["params"], {}
        else:
            raise Exception("Params neither list or dict")

    async def perform(self, content):
        actor = content.get("current_user")
        args, kwargs = self.get_params(content)
        method = content["method"]
        content["thread"] = threading.get_ident()
        if ":" in method:
            mod_name, func_name = method.split(":")
            procedure = getattr(importlib.import_module(mod_name), func_name)
        else:
            procedure = getattr(self.procedures, method)
        try:
            LOGGER.info("%r %s %s", method, args, kwargs)
            result = await perform.perform(actor, procedure, args, kwargs)
            content["result"] = result
            if isinstance(result, perform.Context):
                content["is_context"] = True
        except Exception as ex:
            content["error"] = {"code": -32000, "message": str(ex)}
        return content

    async def _broadcast_(self, signal, message, filter_clients=None):
        msg = {"signal": signal, "message": message, "filter_clients": filter_clients}
        await self.broadcast_redis.publish("broadcast", dumps(msg))

    def broadcast(self, signal, message, filter_clients=None):
        """
            a thread safe way to broadcast
        """
        self.ioloop.spawn_callback(self._broadcast_, signal, message, filter_clients)

    async def do_work(self, val):
        content = loads(val[1].decode("utf-8"))
        if content.get("id"):
            await self._broadcast_("broadcast", {"working": content["id"]})
        content = await self.perform(content)
        if content.get("reply"):
            await self.broadcast_redis.publish(content["reply"], dumps(content))
            LOGGER.info("published reply %s", content)

    async def work(self):
        """ Connects to redis a awaits messages from queue """
        self.broadcast_redis = await connect()
        self.work_q = await connect()
        LOGGER.info("ready for work")
        while True:
            val = await self.work_q.blpop(options.redis_work_q)
            if val:
                await self.do_work(val)
        self.work_q.close()
        await self.work_q.wait_closed()
        self.broadcast_redis.close()
        await self.broadcast_redis.wait_closed()

    async def put(self, content):
        await self.broadcast_redis.execute(
            "rpush", options.redis_work_q, dumps(content)
        )


def main():
    """ Creates Worker and starts tornado """
    asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
    values = settings()
    if options.settings_extend:
        values = extend(options.settings_extend, values)

    assert options.redis_url
    autoreload.start()
    ioloop = IOLoop.current()
    perform._manager_ = Worker(ioloop=ioloop)
    for _ in range(options.workers):
        ioloop.spawn_callback(perform._manager_.work)
    try:
        ioloop.start()
    except KeyboardInterrupt:
        logging.info("stopping")



from tornado.options import options
from tornado.ioloop import IOLoop
from tornado.web import Application
from ..work.service_manager import ServiceManager
from ..work.manager import Manager
from ..utils.config import extend
from .settings import settings
from .urls import urls
import importlib
import logging

LOGGER = logging.getLogger(__name__)

"""
    Simple Tornado Server
"""


def make_app():
    values = settings()

    if options.settings_extend:
        values = extend(options.settings_extend, values)

    pcs = importlib.import_module(options.procedures)

    if options.redis_url:
        from ..work.redis_manger import RedisManager

        values["manager"] = RedisManager(pcs, options.workers)
    else:
        logging.info("local manager")
        values["manager"] = Manager(pcs, options.workers)

    if options.services:
        services = importlib.import_module(options.services)
        services = ServiceManager(services)
        logging.info("services: %s", services.available_services)
        values["manager"].add_targets(services)
        values["services"] = services

    paths = urls()

    if options.urls_extend:
        paths = extend(options.urls_extend, paths)

    return Application(paths, **values)


def main():
    app = make_app()
    app.listen(options.port)
    LOGGER.info("listening on port %s", options.port)

    ioloop = IOLoop.current()
    try:
        ioloop.start()
    except KeyboardInterrupt:
        logging.info("stopping")
        ioloop.stop()



import json
from decimal import Decimal
import datetime
from collections.abc import Callable


try:
    from bson.objectid import ObjectId
except ImportError:
    # in case you're not using Mongodb
    class ObjectId(object):
        pass


class DateTimeEncoder(json.JSONEncoder):
    """
        Encodes datetimes and Decimals
        calls to_json on object if it has that method
    """

    def default(self, obj):
        try:
            if hasattr(obj, "isoformat"):
                return obj.isoformat().replace("T", " ")
            elif isinstance(obj, datetime.date):
                return str(obj)
            elif isinstance(obj, Decimal):
                return float(obj)
            elif hasattr(obj, "to_json") and isinstance(
                getattr(obj, "to_json"), Callable
            ):
                return obj.to_json()
            elif isinstance(obj, ObjectId):
                return str(obj)
            return json.JSONEncoder.default(self, obj)
        except:
            raise


#            return str(obj)


def loads(*args, **kwargs):
    """ calls json.loads """
    return json.loads(*args, **kwargs)


def dumps(o, **kwargs):
    """ calls json.loads with DateTimeEncoder """
    return json.dumps(o, cls=DateTimeEncoder, **kwargs)



from bluemax import context
import time
import asyncio
from tornado import gen
from typing import Tuple


__all__ = ["add", "a_add", "t_add"]


def add(a: int = 2, b: int = 2, sleep_for: float = 0.1) -> Tuple[str, int]:
    """ returns a plus b syncronous."""
    result = a + b
    time.sleep(sleep_for)
    context.broadcast_on_success(
        "action", {"name": "add", "result": result, "user": context.get_current_user()}
    )
    return "add", result


async def a_add(a: int = 2, b: int = 3, sleep_for: float = 0.2) -> Tuple[str, int]:
    """ returns a plus b, async native."""
    result = a + b
    await asyncio.sleep(sleep_for)
    context.broadcast_on_success(
        "action",
        {"name": "a_add", "result": result, "user": context.get_current_user()},
    )
    return "a_add", result


@gen.coroutine
def t_add(a: int = 2, b: int = 4, sleep_for: float = 0.3) -> Tuple[str, int]:
    """ returns a plus b, async tornado."""
    result = a + b
    yield gen.sleep(sleep_for)
    context.broadcast_on_success(
        "action",
        {"name": "t_add", "result": result, "user": context.get_current_user()},
    )
    return "t_add", result



from .manager import Manager
from tornado import gen
from tornado.options import options
from aioredis.pubsub import Receiver
from aioredis.abc import AbcChannel
from ..utils import rd_session
from ..utils import json_utils
from . import context
import aioredis
import asyncio
import logging
import json
import uuid

LOGGER = logging.getLogger(__name__)


class RedisManager(Manager):
    """
        A subclass of manager that delegates
        to a redis queue. It also subscribes
        to broadcasts and sends them on.
    """

    def _setup_queues_(self, workers):
        self._rd_client = {}
        self.broadcast_redis = None
        self.subscribe_redis = None
        self.ioloop.spawn_callback(self.setup_redis)

    async def setup_redis(self):
        loop = asyncio.get_event_loop()
        self.broadcast_redis = await rd_session.connect()
        self.subscribe_redis = await rd_session.connect()
        self.mpsc = Receiver(loop=loop)

        async def reader(mpsc):
            async for channel, msg in self.mpsc.iter():
                assert isinstance(channel, AbcChannel)
                LOGGER.debug("Got %r in channel %r", msg, channel)
                if channel.name == b"broadcast":
                    Manager._broadcast_(self, **json_utils.loads(msg))
                else:
                    client = self._rd_client.get(channel.name.decode("utf-8"))
                    result = json_utils.loads(msg)
                    if result and result.get("is_context") is True:
                        LOGGER.debug("found context")
                        messages = result["result"]["messages"]
                        result["result"] = result["result"]["result"]
                        if client:
                            LOGGER.debug("writing to client %s %r", client.id, result)
                            client.write_message(json_utils.dumps(result))
                        for signal, message, filter_clients in messages:
                            self.broadcast(signal, message, filter_clients)
                    elif client:
                        LOGGER.debug("writing to client %s %r", client.id, result)
                        client.write_message(json_utils.dumps(result))

        asyncio.ensure_future(reader(self.mpsc))
        LOGGER.info("subscribed to redis")
        await self.subscribe_redis.subscribe(self.mpsc.channel("broadcast"))
        LOGGER.info("listening for broadcast")

    def add_client(self, client):
        super().add_client(client)
        self.ioloop.add_callback(self._add_client, client)

    async def _add_client(self, client):
        channel = f"channel:{client.id}"
        self._rd_client[channel] = client
        await self.subscribe_redis.subscribe(self.mpsc.channel(channel))
        LOGGER.info(f"listening for {channel}")

    def remove_client(self, client):
        super().remove_client(client)
        self.ioloop.add_callback(self._remove_client, client)

    async def _remove_client(self, client):
        channel = f"channel:{client.id}"
        await self.subscribe_redis.unsubscribe(channel)
        if channel in self._rd_client.keys():
            del self._rd_client[channel]
        LOGGER.info(f"stopped {channel}")

    async def _broadcast_(self, signal, message, filter_clients=None):
        msg = {"signal": signal, "message": message, "filter_clients": filter_clients}
        LOGGER.info("broadcasting %s", signal)
        if self.broadcast_redis:
            await self.broadcast_redis.publish("broadcast", json_utils.dumps(msg))

    @gen.coroutine
    def put(self, content):
        if content.get("client_id"):
            content["reply"] = "channel:{client_id}".format(**content)
        procedure = self.get_target(content["method"])
        pn = f"{procedure.__module__}:{procedure.__name__}"
        content["method"] = pn
        yield self.broadcast_redis.execute(
            "rpush", options.redis_work_q, json_utils.dumps(content)
        )



from tornado.options import parse_command_line, parse_config_file, options
import os
import logging
import importlib

DEFAULT_FORMAT_STRING = (
    "%(asctime)s %(levelname)-6s: %(module)-11s (%(process)d|%(thread)x): %(message)s"
)


def load_config(path=None):
    """
        This extends the tornado parser to enable use in
        heroku where options are accessed through os.getenv

        Will read file at path if exists

        Will then read environment variables to override

        Will then parse command line to override

    """
    logging.basicConfig(level=logging.INFO, format=DEFAULT_FORMAT_STRING)

    if path is not None and os.path.isfile(path):
        parse_config_file(path)
        logging.info("loaded config from %s", path)

    for k in options.as_dict():
        """ danger: access of private variables """
        value = os.getenv(k)
        if value:
            name = options._normalize_name(k)
            option = options._options.get(name)
            option.parse(value)

    parse_command_line()


def extend(extension, value):
    mdl, fn = extension.split(":")
    fn = getattr(importlib.import_module(mdl), fn)
    return fn(value)



from bluemax.utils.config import load_config
from tornado.options import define, options

define("debug", type=bool, default=True, help="auto reload etc")
define("port", type=int, default=8080, help="port to listen on")
define("workers", type=int, default=1, help="number of worker threads")
define("procedures", type=str, default="", help="procedures to make available")
define("services", type=str, default="", help="services to make available")
define("static_dir", type=str, default=None, help="static web directory")
define("redis_url", default="", help="redis url")
define("redis_work_q", default="_work_q_", help="name of queue to rpush to")

define("cookie_name", type=str, default="bluemax", help="bluemax session cooklie name")
define(
    "cookie_secret",
    type=str,
    default="it was a dark and stormy night, blue max",
    help="session secret",
)
define("login_url", type=str, default="/login", help="url for login")
define("auth_class", type=str, default=None, help="class of login handlers")

define("ms_app_id", type=str, help="MS App ID for authentication")
define("ms_app_secret", type=str, help="MS App Secret for authentication")

define("settings_extend", type=str, help="fully qualified function to import and call")
define("urls_extend", type=str, help="fully qualified function to import and call")


def settings():
    result = {
        "cookie_secret": options.cookie_secret,
        "login_url": options.login_url,
        "debug": options.debug,
    }
    return result



import logging
from bluemax.scripts.pid_script import Stoppable


logger = logging.getLogger(__name__)


class Worker(Stoppable):
    def start(self):
        """
            Called as console_script to run the
            worker using redis.
        """
        from bluemax.web.settings import load_config, options
        from bluemax.work import worker

        load_config(".env")
        logger.info("hello from worker! mode:%s", "debug" if options.debug else "prod")
        worker.main()


def main():
    Worker().main()


if __name__ == "__main__":
    main()



from bluemax.web.settings import load_config
from tornado.options import options
from ..web import settings
from ..web import urls
import click
import signal
import os
import sys
import importlib
import logging

sys.path.insert(0, os.getcwd())

logger = logging.getLogger(__name__)


def write_pid(path):
    pid = os.getpid()
    with open(path, "w") as f:
        f.write(f"{pid}")


def import_module(name):
    "Íts ok not to have the module but not ok to not have dependants."
    try:
        return importlib.import_module(name)
    except ModuleNotFoundError as ex:
        if name not in str(ex):
            raise


def get_module(name):
    if name is None:
        print("module expected")
    else:
        md = importlib.import_module(name)
        procs = importlib.import_module(f"{name}.procedures")
        if procs:
            if not getattr(procs, "__all__"):
                print(f"expected __all__ in procedures module")
            else:
                options["procedures"] = f"{name}.procedures"
                ext_settings = import_module(f"{name}.settings")
                if ext_settings:
                    options["settings_extend"] = f"{name}.settings:extend"
                ext_urls = import_module(f"{name}.urls")
                if ext_urls:
                    options["urls_extend"] = f"{name}.urls:extend"
                return True
        else:
            print(f"expected procedures in {name}")


@click.group()
def cli():
    pass


@cli.command()
@click.option("-p", "--pid", default=None, help="/path/to/pid/file")
@click.option("-m", "--module", required=True, help="module to remote")
@click.option("-s", "--services", required=False, help="services module")
def server(pid, module, services):
    from bluemax.web import server as m_server

    if pid is None:
        pid = "server.pid"
    if get_module(module):
        if services:
            options.services = services
        load_config(".env")
        logger.info("mode: %s", "debug" if options.debug else "prod")
        write_pid(pid)
        m_server.main()


@cli.command()
@click.option("-p", "--pid", default=None, help="/path/to/pid/file")
@click.option("-m", "--module", required=True, help="module to remote")
def worker(pid, module):
    from bluemax.work import worker as m_worker

    if pid is None:
        pid = "worker.pid"
    if get_module(module):
        load_config(".env")
        logger.info(
            "hello from worker! mode: %s", "debug" if options.debug else "prod"
        )
        write_pid(pid)
        m_worker.main()


@cli.command()
@click.argument("pid_path", type=click.Path(exists=True))
def stop(pid_path):
    try:
        with open(pid_path, "r") as f:
            contents = f.read()
        pid = int(contents)
        os.kill(pid, signal.SIGINT)
        os.remove(pid_path)
    except:
        raise Exception("could not find %s", pid_path)


@cli.command()
@click.argument("name")
@click.option("-f", "--force", is_flag=True, default=False, help="force new project")
def startproject(name, force):
    print(f"gen {name}")
    if os.path.exists(name) and force is False:
        print("that module already exists: ", name)
    else:
        if not os.path.isdir(name):
            os.makedirs(name)
        with open(os.path.join(name, "__init__.py"), "w") as file:
            file.write("")
        with open(os.path.join(name, "settings.py"), "w") as file:
            file.write(
                """''' add your settings here '''
import logging

def extend(settings):
    logging.info('extending settings')
    return settings
"""
            )
        with open(os.path.join(name, "urls.py"), "w") as file:
            file.write(
                """''' add your tornado routes here '''
import logging

def extend(urls):
    logging.info('extending urls')
    return urls
"""
            )
        with open(os.path.join(name, "procedures.py"), "w") as file:
            file.write(
                """''' an example function and exposing it through __all__ '''

__all__=['add']

def add(a:int, b:int)->int:
    return a+b
"""
            )
        with open(os.path.join(name, "services.py"), "w") as file:
            file.write(
                """''' an example service '''
import asyncio
import time
from bluemax import context

async def clock():
    ''' Will broadcast every 5 seconds '''
    while True:
        context.broadcast('time', {'now': time.time()})
        await asyncio.sleep(5)
"""
            )



class DotDict(dict):
    """
        A dict that allows for object-like property access syntax.
    """

    def __getattr__(self, name):
        if name[0] == "_":
            return dict.__getattr__(self.name)
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)



from .perform import (
    get_current_user,
    broadcast,
    broadcast_on_success,
    put_work,
    register_object,
    deregister_object,
)

__all__ = [
    "get_current_user",
    "broadcast",
    "broadcast_on_success",
    "put_work",
    "register_object",
    "deregister_object",
]



from pkg_resources import resource_filename
from tornado.web import StaticFileHandler
from tornado.options import options
from .authentication.auth_static_file_handler import AuthStaticFileHandler
from .authentication import LogoutHandler
from .json_rpc2 import JsonRpcHandler
import importlib
import logging


def urls():
    static_dir = options.static_dir
    if static_dir is None:
        static_dir = resource_filename("bluemax.web", "static")
    logging.info("static_dir %s", static_dir)
    if options.auth_class:
        FileHandler = AuthStaticFileHandler
    else:
        FileHandler = StaticFileHandler
    result = [
        (r"/rpc", JsonRpcHandler),
        (r"/(.*)", FileHandler, {"path": static_dir, "default_filename": "index.html"}),
    ]
    if options.auth_class:
        module_name, class_name = options.auth_class.split(":")
        login_handler = getattr(importlib.import_module(module_name), class_name)
        result = [(r"/login", login_handler), (r"/logout", LogoutHandler)] + result
    return result



from sqlalchemy.orm import class_mapper, ColumnProperty
from sqlalchemy.orm.session import object_session
from collections import OrderedDict
import logging


def dumps(item):
    result = {"_type_": item.__class__.__name__}
    for prop in class_mapper(item.__class__).iterate_properties:
        if prop.key[0] == "_":
            continue
        if isinstance(prop, ColumnProperty):
            result[prop.key] = getattr(item, prop.key)
    return result


def loads(values, item):
    for prop in class_mapper(item.__class__).iterate_properties:
        if prop.key[0] == "_":
            continue
        if isinstance(prop, ColumnProperty) and prop.key in values.keys():
            setattr(item, prop.key, values.get(prop.key))


def from_pairs(o, obj):
    for key in o.__mapper__.c.keys():
        if key[0] == "_" or key in ["id", "last_updated"]:
            continue
        value = obj.get(key)
        logging.debug("set %s:%r", key, value)
        if key == "version_id" and value == None:
            continue
        setattr(o, key, value)


def from_one_to_many_pairs(o, pairs, key, clazz):
    """Simple saving of one-to-many relations - not usually associated with from_pairs"""
    added = []
    updated = []
    deleted = []
    errors = []
    session = object_session(o)
    existing_keys = [i.id for i in getattr(o, key)]
    for item in pairs.get(key, []):
        item_id = item.get("id")
        obj = None
        if item_id:
            if int(item_id) < 0:
                obj = session.query(clazz).get(abs(int(item_id)))
                session.delete(obj)
                deleted.append(obj)
                continue
            else:
                if item_id in existing_keys:
                    existing_keys.remove(item_id)
                obj = session.query(clazz).get(item_id)
                if from_pairs(obj, item) is True:
                    e = obj._validate_()
                    if e:
                        errors.extend(e)
                    updated.append(obj)
        else:
            obj = clazz()
            getattr(o, key).append(obj)
            if from_pairs(obj, item) is True:
                e = obj._validate_()
                if e:
                    errors.extend(e)
                added.append(obj)
    for key in existing_keys:
        obj = session.query(clazz).get(abs(int(key)))
        session.delete(obj)
        deleted.append(obj)
    return added, updated, deleted, errors


def from_many_to_many_pairs(o, pairs, key, clazz):
    """Simple saving of many-to-many relations - not usually associated with from_pairs"""
    added = []
    removed = []
    session = object_session(o)
    existing = dict([(a.id, a) for a in getattr(o, key)])
    new_list = dict([(b.get("id"), b) for b in pairs.get(key, [])])
    for to_remove in set(existing.keys()) - set(new_list.keys()):
        obj = existing[to_remove]
        getattr(o, key).remove(obj)
        removed.append(obj)
        logging.info("m2m removed %s", obj)
    for to_add in set(new_list.keys()) - set(existing.keys()):
        obj = session.query(clazz).get(abs(int(to_add)))
        getattr(o, key).append(obj)
        added.append(obj)
        logging.info("m2m added %s", obj)
    return added, removed



from setuptools import setup, find_packages

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = True


except ImportError:
    bdist_wheel = None


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="bluemax",
    version="0.0.47",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/blueshed/bluemax/",
    packages=find_packages(exclude=["tests.*", "tests"]),
    package_data={"bluemax.web": ["static/*"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["Click", "tornado"],
    extras_require={
        "sa": ["alembic", "sqlalchemy"],
        "redis": ["aioredis", "redis"],
        "all": ["alembic", "sqlalchemy", "aioredis", "redis"],
        "dev": ["pytest-tornado", "wheel", "twine", "bumpversion", "invoke"],
    },
    entry_points={"console_scripts": [
        "runmax=bluemax.scripts.runmax:cli"
    ]},
)



from tornado.web import authenticated, StaticFileHandler
from .authentication import UserMixin


class AuthStaticFileHandler(UserMixin, StaticFileHandler):
    """
    This provide integration between tornado.web.authenticated
    and tornado.web.StaticFileHandler.

    It assumes you have set up the cookie name in the application
    settings and that the request already has the cookie set. In
    other words the user has already authenticated.
    """

    @authenticated
    def get(self, path, include_body=True):
        return StaticFileHandler.get(self, path, include_body)



from tornado.websocket import WebSocketHandler
from tornado.web import HTTPError
from tornado import gen
from .authentication import UserMixin
from ..utils import json_utils
import functools
import logging
import uuid

"""
    Implentation of the json rpc 2.0 spec.
"""

LOGGER = logging.getLogger(__name__)


class JsonRpcException(Exception):
    def __init__(self, code, message):
        super().__init__(self, message)
        self.js_code = code
        self.js_message = message


class JsonRpcHandler(UserMixin, WebSocketHandler):
    def initialize(self):
        self.id = str(uuid.uuid4())

    @property
    def manager(self):
        return self.settings["manager"]

    @property
    def procedures(self):
        return self.manager.proc_names

    def open(self):
        user = self.get_current_user()
        if user:
            self.manager.add_client(self)
            LOGGER.info("user: %s", user)
            self.write_message(
                json_utils.dumps({"signal": "set_user", "message": user})
            )
        elif self.settings["debug"] is False:
            raise HTTPError(403)
        else:
            self.manager.add_client(self)

    def get_params(self, content):
        if isinstance(content["params"], dict):
            return [], content["params"]
        elif isinstance(content["params"], list):
            return content["params"], {}
        else:
            raise JsonRpcException(-32602, "Params neither list or dict")

    @gen.coroutine
    def perform(self, content, result):
        method = content["method"]
        if method == "authenticate":
            result["result"] = self.get_transfer_user(content["params"][0])
            self.write_message(json_utils.dumps(result))
        else:
            content["client_id"] = self.id
            content["current_user"] = self.get_current_user()
            future = yield self.manager.put(content)
            if future:
                future.add_done_callback(functools.partial(self.work_done, result))

    @gen.coroutine
    def on_message(self, data):
        content = json_utils.loads(data)
        LOGGER.info(content)
        result = {"jsonrpc": "2.0"}
        try:
            if content.get("jsonrpc") != "2.0":
                raise JsonRpcException(-32600, "protocol not supported")
            ref = content.get("id")
            if ref:
                result["id"] = ref

            method = content.get("method")
            if method is None:
                raise JsonRpcException(-32600, "no method")
            if method[0] == "_":
                raise JsonRpcException(-32600, "method private")
            if method in self.procedures + ["authenticate"]:
                yield self.perform(content, result)
                result = None
            else:
                raise JsonRpcException(-32600, f"no such method: {method}")

            if ref is None:
                result = None  # its a notification

        except JsonRpcException as ex:
            LOGGER.exception(ex)
            result["error"] = {"code": ex.js_code, "message": ex.js_message}
        except Exception as ex:
            LOGGER.exception(ex)
            result["error"] = {"code": -32000, "message": str(ex)}
        finally:
            if result:
                LOGGER.debug(repr(result))
                self.write_message(json_utils.dumps(result))

    def work_done(self, result, future):
        try:
            output = future.result()
            result["result"] = output
        except Exception as ex:
            result["error"] = {"code": -32000, "message": str(ex)}
        self.write_message(json_utils.dumps(result))

    def on_close(self):
        self.manager.remove_client(self)
